#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <assert.h>

#include "tparser.h"
#include "parser.h"

int debug;

extern char *yytext;

extern pt_script script;

static FILE *file = NULL;
static int eof = 0;
static int nRow = 0;
static int nBuffer = 0;
static int lBuffer = 0;
static int nTokenStart = 0;
static int nTokenLength = 0;
static int nTokenNextStart = 0;
static int lMaxBuffer = PARSER_MAX_LINE;
static char buffer[PARSER_MAX_LINE+1];

char panoParserDumpChar(char c) 
{
    if (  isprint(c)  )
	return c;
    return '@';
}

void panoParserDumpRow(void) 
{
    fprintf(stdout, "%6d |%.*s", nRow, lBuffer, buffer);
}

void panoParserError(char const *errorstring, ...) 
{
    static char errmsg[1000];
    va_list args;
    
    int start=nTokenStart;
    int end=start + nTokenLength - 1;
    int i;
    
    panoParserDumpRow();
    
    if (  eof  ) {
	fprintf(stdout, "       !");
	for (i=0; i<lBuffer; i++)
	    fprintf(stdout, ".");
	fprintf(stdout, "^-EOF\n");
    } else {
	fprintf(stdout, "       !");
	for (i=1; i<start; i++)
	    fprintf(stdout, ".");
	for (i=start; i<=end; i++)
	    fprintf(stdout, "^");
	fprintf(stdout, "   at line %d column %d\n", nRow, start);
    }

    /* print it using variable arguments -----------------------------*/
    va_start(args, errorstring);
    vsprintf(errmsg, errorstring, args);
    va_end(args);
    
    fprintf(stdout, "Error: %s\n", errmsg);

    // exit
}


int getNextLine(void) 
{
  char *p;
  
  nBuffer = 0;
  nTokenStart = -1;
  nTokenNextStart = 1;
  eof = FALSE;

  p = fgets(buffer, lMaxBuffer, file);
  if (  p == NULL  ) {
    if (  ferror(file)  )
      return -1;
    eof = TRUE;
    return 1;
  }

  nRow += 1;
  lBuffer = strlen(buffer);

  return 0;
}

// THis is the function that lex will use to read the next character
int GetNextChar(char *b, int maxBuffer) 
{
  int frc;
  
  if (  eof  )
    return 0;
  
  // read next line if at the end of the current
  while (  nBuffer >= lBuffer  ) {
    frc = getNextLine();
    if (  frc != 0  )
      return 0;
    }

  // ok, return character 
  b[0] = buffer[nBuffer];
  nBuffer += 1;

  if (  debug  )
    printf("GetNextChar() => '%c'0x%02x at %d\n",
                        panoParserDumpChar(b[0]), b[0], nBuffer);
  return b[0]==0?0:1;
}

void panoScannerTokenBegin(char *t) 
{
    // REcord where a token begins
    nTokenStart = nTokenNextStart;
    nTokenLength = strlen(t);
    nTokenNextStart = nBuffer; // + 1;
}


void yyerror (char const *st)
{
    panoParserError(st);
}

void panoParserDumpImage(pt_script_image *output)
{
    printf("   projection %d\n", output->projection);
    printf("   width, height %dx%d ", output->width, output->height);
    printf("   fHorFOV %f\n", output->fHorFOV);
    printf("   yaw %f ", output->yaw);
    printf("   pitch %f ", output->pitch);
    printf("   roll %f ", output->roll);
    
    printf("   a %f", output->coef[0]);
    printf("   b %f", output->coef[1]);
    printf("   c %f", output->coef[2]);
    printf("   d %f", output->coef[3]);
    printf("   e %f", output->coef[4]);
    printf("   g %f", output->coef[5]);
    printf("   t %f\n", output->coef[6]);
    
    printf("   name %s\n", output->name);
    printf("   cropType %d ", output->cropType);
    printf("   cropArea %i,%i,%i,%i ", 
	   output->cropArea[0],
	   output->cropArea[1],
	   output->cropArea[2],
	   output->cropArea[3]);
    printf("   morphToFit %d\n", output->morphToFit);

    printf("   References: ");
    printf("   fHorFOV %d", output->fHorFOVIndex);
    printf("   yaw  %d", output->yawIndex);
    printf("   pitch  %d", output->pitchIndex);
    printf("   roll  %d", output->rollIndex);
    
    printf("   a  %d", output->coefIndex[0]);
    printf("   b  %d", output->coefIndex[1]);
    printf("   c  %d", output->coefIndex[2]);
    printf("   d  %d", output->coefIndex[3]);
    printf("   e  %d", output->coefIndex[4]);
    printf("   g  %d", output->coefIndex[5]);
    printf("   t  %d\n", output->coefIndex[6]);
    
}

int panoParseReset(void)
{
    if (file != NULL) 
	return FALSE;
    /* There should not be anything allocated in script */
    bzero(&script, sizeof(script));    
    return 1;
}


pt_script *panoParseScript(char *filename)
{
    pt_script *scriptVar;
    if (!panoParseReset() ) {
	// This is really an assertion
	fprintf(stderr, "This parser is not reentrant");
	exit(1);
    }
    file = fopen(filename, "r");
    if (  file == NULL  ) {
	return NULL;
    }
    if ( getNextLine() == 0 ) {
	if (yyparse() ==0) {
	    // XXX
	    // duplicate data structure to return
	    scriptVar = calloc(1, sizeof(*scriptVar));
	    if (scriptVar == NULL) {
		PrintError("Not enough memory");
		return NULL;
	    }
	    // Just making sure..
	    assert(sizeof(*scriptVar) == sizeof(*scriptVar));
	    memcpy(scriptVar, &script, sizeof(*scriptVar));
	    return scriptVar;
	} else  {
	    return FALSE;
	}
	    
    }
    else  {
	PrintError("Input file is empty");
	return NULL;
    }
    close (file);
    
}

void panoParseDispose(pt_script **scriptVar)
{
    int i;
    // free all the data structures in it before freeing the variable itself

#define FREE(a) ((a) != NULL, free(a))
    pt_script *ptr = *scriptVar;
    
    // Free pt_script_pano first
    FREE(ptr->pano.projectionParms);
    
    FREE(ptr->pano.outputFormat);

    for (i=0;i<ptr->iOutputImagesCount;i++) {
	FREE(ptr->outputImageSpec[i].name);
    }
    for (i=0;i<ptr->iInputImagesCount;i++) {
	FREE(ptr->inputImageSpec[i].name);
    }
    FREE(ptr->outputImageSpec);
    FREE(ptr->inputImageSpec);
    FREE(ptr->morphingPointsSpec);

#undef FREE    

    free(ptr);
    *scriptVar = NULL;
}



int main(int argc, char *argv[]) 
{
    int i;

    pt_script_image *output;
    pt_script *script = NULL;

    script = panoParseScript(argv[1]);

    if (script == NULL) {
	printf("Parsing error\n");
	return 1;
    }

    
    printf("Values\n");
    printf("Pano .................");
    printf(" width,height %dx%d ", script->pano.width, script->pano.height);
    printf(" projection %d ", script->pano.projection);
    printf(" name %s ", script->pano.projectionName);
    printf(" fHorFOV %f ", script->pano.fHorFOV);
    printf(" format %s ", script->pano.outputFormat);
    printf(" proj parms %s\n", script->pano.projectionParms);
    
    printf("Optimize ..................");
    printf("  gamma %f", script->optimize.fGamma);
    printf("  interpolator %d", script->optimize.interpolator);
    printf("  fastFT %d ", script->optimize.fastFT);
    printf("  humber %d \n", script->optimize.humberEstimator);
    
    printf("Output  [%d] images ..................\n", script->iOutputImagesCount);
    
    for (i=0;i<script->iOutputImagesCount;i++) {
	output = &script->outputImageSpec[i];
	printf(" image %d\n", i);
	panoParserDumpImage(output);
    }
    printf("Input  [%d] images ..................\n", script->iInputImagesCount);

    for (i=0;i<script->iInputImagesCount;i++) {
	output = &script->inputImageSpec[i];
	printf(" image %d\n", i);
	panoParserDumpImage(output);
    }

    panoParseDispose(&script);

    return 0;
}
