/*
 *  parser.y
 *
 *  Copyright  Daniel M. German
 *  
 *  April 2007
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  Author: Daniel M German dmgerman at uvic doooot ca
 * 
 */

%{
       #include <math.h>
       #include <stdio.h>
       #include <stdlib.h>
       #include <string.h>

#include "tparser.h"

       int yylex (void);
       void yyerror (char const *);
/*  Keeps track of the current type of input line in the file */
int  currentLine = -1;

pt_script script;

/* defining it gives better error messages. It might be an overkill */
#define YYERROR_VERBOSE 1 

static pt_script_image *image;


/* copy a string while allocating and checking for memory */
static void ParserStringCopy(char **dest, char *from)
{
    if (*dest) 
	free(*dest);
    *dest = strdup(from);
    if (*dest == NULL)
	panoParserError("Not enough memory");
}


%}
     
%defines

%union {
  int   iVal;
  float fVal;
  char strVal[PT_TOKEN_MAX_LEN+1];
  char cVal;
}


%token NUM
%token PT_TOKEN_SEP
%token <iVal> PT_TOKEN_INT
%token <fVal> PT_TOKEN_FLOAT
%token <strVal> PT_TOKEN_STRING
%token <cVal> PT_TOKEN_KEYWORD
%token PT_TOKEN_EOL
%token PT_TOKEN_SEPARATOR
%token PT_TOKEN_INPUT_LINE
%token PT_TOKEN_OUTPUT_LINE
%token PT_TOKEN_PANO_LINE
%token PT_TOKEN_OPTIMIZE_LINE
%token PT_TOKEN_CONTROL_PTS_LINE
%token PT_TOKEN_COMMA
%token PT_TOKEN_REFERENCE


%start input

%% /* Grammar rules and actions follow.  */

input: lines 
      { return 0;}

lines: 
        | line input
{
    printf("Per line...\n");
}
     
line:     eoln { ; }
             | inputline
             | outputline
             | panoline
             | optimizeline
             | PT_TOKEN_CONTROL_PTS_LINE PT_TOKEN_SEPARATOR vars eoln      { ; }
;
     
eoln:     PT_TOKEN_EOL  { currentLine = -1; /* This says we don't know the type of line being processed */}

inputline:  PT_TOKEN_INPUT_LINE PT_TOKEN_SEPARATOR
        {
	    currentLine = PT_TOKEN_INPUT_LINE; 
	    // allocate the new output
	    script.iInputImagesCount++;
	    script.inputImageSpec = realloc(script.inputImageSpec, 
					     sizeof(*script.inputImageSpec) * script.iInputImagesCount);
	    if (script.inputImageSpec == NULL) {
		yyerror("Not enough memory");
	    }
	    image = &script.inputImageSpec[script.iInputImagesCount-1];
        }   
        vars eoln      { ; }

outputline: PT_TOKEN_OUTPUT_LINE PT_TOKEN_SEPARATOR 
        {
	    currentLine = PT_TOKEN_OUTPUT_LINE; 
	    // allocate the new output
	    script.iOutputImagesCount++;
	    script.outputImageSpec = realloc(script.outputImageSpec, 
					     sizeof(*script.outputImageSpec) * script.iOutputImagesCount);
	    if (script.outputImageSpec == NULL) {
		yyerror("Not enough memory");
	    }
	    image = &script.outputImageSpec[script.iOutputImagesCount-1];
	}      vars eoln      { ; }

optimizeline: PT_TOKEN_OPTIMIZE_LINE PT_TOKEN_SEPARATOR { currentLine=PT_TOKEN_OPTIMIZE_LINE; } vars eoln      { ; }

panoline: PT_TOKEN_PANO_LINE  PT_TOKEN_SEPARATOR { currentLine=PT_TOKEN_PANO_LINE; } vars eoln      { ; }


vars:      var
          | vars PT_TOKEN_SEPARATOR var

var:       varparameter | varseveralparms | varreference


 /*  Rule for input image field references <var>=<index> */
varreference: PT_TOKEN_KEYWORD PT_TOKEN_REFERENCE PT_TOKEN_INT
   {
       int imageRef = $3 + 1;
       switch (currentLine) {
       case PT_TOKEN_INPUT_LINE:
	   switch ($1) {
	    case 'v':
		image->fHorFOVIndex = imageRef;
		break;
	    case 'y':
		image->yawIndex = imageRef;
		break;
	    case 'p':
		image->pitchIndex = imageRef;
		break;
	    case 'r':
		image->rollIndex = imageRef;
		break;
	    case 'a':
	    case 'b':
	    case 'c':
	    case 'd':
	    case 'e':
		image->coefIndex[$1-'a'] = imageRef;
		break;
	    case 'g':
		image->coefIndex[5] = imageRef;
		break;
	    case 't':
		image->coefIndex[6] = imageRef;
		break;
	   default:
	       panoParserError("Invalid variable name [%c] in input line\n", $1);
	       return -1;
	       break;
	   }
	   break;
       default:
	   panoParserError("Error Not handled 3 [%c]\n", $1);
	   return -1;
       }
   }

 /* Rule for [CS]<x>,<x>,<x>,<x> */
varseveralparms: PT_TOKEN_KEYWORD PT_TOKEN_INT PT_TOKEN_COMMA PT_TOKEN_INT PT_TOKEN_COMMA PT_TOKEN_INT PT_TOKEN_COMMA PT_TOKEN_INT 
    {
	switch (currentLine) {
	case PT_TOKEN_OUTPUT_LINE:
	case PT_TOKEN_INPUT_LINE:
	    switch ($1) {
	    case 'C':
	    case 'S':
		image->cropType = $1;
		image->cropArea[0] = $2;
		image->cropArea[1] = $4;
		image->cropArea[2] = $6;
		image->cropArea[3] = $8;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in image line\n", $1);
		return -1;
	    }
	    break;
	default:
	    panoParserError("Error Not handled 3\n");
	    return -1;
	}
    }

/*  
    Rules for <variable><parameter> 
    I repeat the processing of some variables in int and float type because I don't really know 
    how to deal with both data types in a simple rule as I don't want to always upgrade ints to floats.
 */
varparameter: PT_TOKEN_KEYWORD PT_TOKEN_STRING 
    { 
        /* Processing of string variables */
	switch (currentLine) {
	case PT_TOKEN_PANO_LINE:
	    switch ($1) {
	    case 'f':
		ParserStringCopy(&script.pano.projectionName, $2);
		break;
	    case 'n':
		ParserStringCopy(&script.pano.outputFormat, $2);
		break;
	    case 'P':
		ParserStringCopy(&script.pano.projectionParms, $2);
		break;
	    default:
		panoParserError("Invalid variable name [%c] in pano line\n", $1);
		return -1;
	    }
	    break;
	case PT_TOKEN_OUTPUT_LINE:
	case PT_TOKEN_INPUT_LINE:
	    switch ($1) {
	    case 'n':
		ParserStringCopy(&image->name, $2);
		break;
	    default:
		panoParserError("Invalid variable name [%c] in image line\n", $1);
		return -1;
	    }
	    break;
	default:
	    panoParserError("Error Not handled case [%c]\n", $1);
	    return -1;
	}
    }
   | PT_TOKEN_KEYWORD PT_TOKEN_INT
    {
        /* Processing of int variables */
	switch (currentLine) {
	case PT_TOKEN_PANO_LINE:
	    switch ($1) {
	    case 'w':
		script.pano.width = $2;
		break;
	    case 'h':
		script.pano.height = $2;
		break;
	    case 'f':
		script.pano.projection = $2;
		break;
	    case 'v':
		script.pano.fHorFOV = $2;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in pano line\n", $1);
		return -1;
	    }
	    break;
	case PT_TOKEN_OUTPUT_LINE:
	case PT_TOKEN_INPUT_LINE:
	    switch ($1) {
	    case 'w':
		image->width = $2;
		break;
	    case 'h':
		image->height = $2;
		break;
	    case 'f':
		image->projection = $2;
		break;
	    case 'v':
		image->fHorFOV = $2;
		break;
	    case 'y':
		image->yaw = $2;
		break;
	    case 'p':
		image->pitch = $2;
		break;
	    case 'r':
		image->roll = $2;
		break;
	    case 'a':
	    case 'b':
	    case 'c':
	    case 'd':
	    case 'e':
		image->coef[$1-'a'] = $2;
		break;
	    case 'g':
		image->coef[5] = $2;
		break;
	    case 't':
		image->coef[6] = $2;
		break;
	    case 'u':
	    case 'm':
		printf("Option %c in image line deprecated. \n", $1);
		break;
	    default:
		panoParserError("Invalid variable name [%c] in image line\n", $1);
		return -1;
	    }
	    break;
	case PT_TOKEN_OPTIMIZE_LINE:
	    switch ($1) {
	    case 'g':
		script.optimize.fGamma = $2;
		break;
	    case 'i':
		script.optimize.interpolator = $2;
		break;
	    case 'f':
		script.optimize.fastFT = $2;
		break;
	    case 'm':
		script.optimize.humberEstimator = $2;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in optimize line\n", $1);
		return -1;
	    }
	    break;
	default:
	    panoParserError("Error. Not handled (token int [%c]\n", $1);
	    return -1;
	}
    }
   | PT_TOKEN_KEYWORD PT_TOKEN_FLOAT  
    { 
        /* Processing of float variables */
	switch (currentLine) {
	case PT_TOKEN_PANO_LINE:
	    switch ($1) {
	    case 'v':
		script.pano.fHorFOV = $2;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in output line\n", $1);
		return -1;
	    }
	    break;
	case PT_TOKEN_OUTPUT_LINE:
	case PT_TOKEN_INPUT_LINE:
	    switch ($1) {
	    case 'y':
		image->yaw = $2;
		break;
	    case 'p':
		image->pitch = $2;
		break;
	    case 'r':
		image->roll = $2;
		break;
	    case 'v':
		image->fHorFOV = $2;
		break;
	    case 'a':
	    case 'b':
	    case 'c':
	    case 'd':
	    case 'e':
		image->coef[$1-'a'] = $2;
		break;
	    case 'g':
		image->coef[5] = $2;
		break;
	    case 't':
		image->coef[6] = $2;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in pano line\n", $1);
		return -1;
	    }
	    break;
	case PT_TOKEN_OPTIMIZE_LINE:
	    switch ($1) {
	    case 'g':
		script.optimize.fGamma = $2;
		break;
	    default:
		panoParserError("Invalid variable name [%c] in optimize line\n", $1);
		return -1;
	    }
	    break;
	default:
	    printf("Error Not handled 2 [%c]\n", $1);
	    return -1;
	}
    }

             

%%
